package ru.dbotthepony.mc.otm.block.entity;

import net.minecraft.MethodsReturnNonnullByDefault;
import net.minecraft.core.BlockPos;
import net.minecraft.core.Direction;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.network.chat.Component;
import net.minecraft.network.chat.TranslatableComponent;
import net.minecraft.server.level.ServerLevel;
import net.minecraft.world.entity.player.Inventory;
import net.minecraft.world.entity.player.Player;
import net.minecraft.world.inventory.AbstractContainerMenu;
import net.minecraft.world.item.ItemStack;
import net.minecraft.world.level.Level;
import net.minecraft.world.level.block.Block;
import net.minecraft.world.level.block.entity.BlockEntity;
import net.minecraft.world.level.block.state.BlockState;
import net.minecraftforge.common.capabilities.Capability;
import net.minecraftforge.common.util.LazyOptional;
import net.minecraftforge.energy.CapabilityEnergy;
import net.minecraftforge.energy.IEnergyStorage;
import net.minecraftforge.items.CapabilityItemHandler;
import net.minecraftforge.items.IItemHandler;
import ru.dbotthepony.mc.otm.Registry;
import ru.dbotthepony.mc.otm.block.BlockBatteryBank;
import ru.dbotthepony.mc.otm.block.BlockMatteryRotatable;
import ru.dbotthepony.mc.otm.capability.IMatteryEnergyStorage;
import ru.dbotthepony.mc.otm.capability.MatteryCapability;
import ru.dbotthepony.mc.otm.container.MatteryContainer;
import ru.dbotthepony.mc.otm.core.Fraction;
import ru.dbotthepony.mc.otm.menu.BatteryBankMenu;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.ParametersAreNonnullByDefault;
import java.math.BigDecimal;
import java.util.Optional;

@MethodsReturnNonnullByDefault
@ParametersAreNonnullByDefault
public class BlockEntityBatteryBank extends BlockEntityMattery {
	// 6 на 2
	public final MatteryContainer battery_container = new MatteryContainer(this::setChanged, 6 * 2) {
		@Override
		public void setChanged(int slot, ItemStack new_state, ItemStack old_state) {
			super.setChanged(slot, new_state, old_state);

			if (level != null) {
				var state = getBlockState();

				for (int i = 0; i < BlockBatteryBank.BATTERY_SLOTS_PROPS.length; i++) {
					state = state.setValue(BlockBatteryBank.BATTERY_SLOTS_PROPS[i], getItem(i).getCapability(CapabilityEnergy.ENERGY).isPresent());
				}

				if (state != getBlockState()) {
					level.setBlock(getBlockPos(), state, Block.UPDATE_CLIENTS);
				}
			}
		}
	};

	private final LazyOptional<IItemHandler> item_handler_resolver = LazyOptional.of(() -> battery_container.handler(
		(slot, stack) -> stack.getCapability(CapabilityEnergy.ENERGY).isPresent(),
		(slot, amount, stack) -> true
	));

	public record BatteryBankDistribution(Fraction[] distribution, Fraction max_throughput) {

	}

	public record BatteryBankEnergy(BlockEntityBatteryBank tile, BankMode mode) implements IMatteryEnergyStorage {
		public enum BankMode {
			RECEIVE,
			EXTRACT,
			BIDIRECTIONAL
		}

		@Override
		public boolean canExtract() {
			return mode == BankMode.EXTRACT || mode == BankMode.BIDIRECTIONAL;
		}

		@Override
		public boolean canReceive() {
			return mode == BankMode.RECEIVE || mode == BankMode.BIDIRECTIONAL;
		}

		@Nonnull
		@Override
		public Fraction extractEnergyOuter(Fraction howMuch, boolean simulate) {
			if (mode == BankMode.RECEIVE)
				return Fraction.ZERO;

			return extractEnergyInner(howMuch, simulate);
		}

		BatteryBankDistribution getDistribution(boolean mode) {
			Fraction[] distribution = new Fraction[tile.battery_container.getContainerSize()];
			Fraction summ = Fraction.ZERO;

			for (int i = 0; i < tile.battery_container.getContainerSize(); i++) {
				ItemStack stack = tile.battery_container.getItem(i);

				if (!stack.isEmpty()) {
					Optional<IMatteryEnergyStorage> cap = stack.getCapability(MatteryCapability.ENERGY).resolve();

					if (cap.isPresent()) {
						Fraction diff = mode ? cap.get().receiveEnergyOuter(MatteryCapability.LONG_MAX_VALUE, true) : cap.get().extractEnergyOuter(MatteryCapability.LONG_MAX_VALUE, true);
						distribution[i] = diff;
						summ = summ.plus(diff);
					} else {
						Optional<IEnergyStorage> cap2 = stack.getCapability(CapabilityEnergy.ENERGY).resolve();

						if (cap2.isPresent()) {
							Fraction diff = new Fraction(mode ? cap2.get().receiveEnergy(Integer.MAX_VALUE, true) : cap2.get().extractEnergy(Integer.MAX_VALUE, true));
							distribution[i] = diff;
							summ = summ.plus(diff);
						}
					}
				} else {
					distribution[i] = Fraction.ZERO;
				}
			}

			if (summ.compareTo(Fraction.ZERO) != 0) {
				for (int i = 0; i < tile.battery_container.getContainerSize(); i++) {
					distribution[i] = distribution[i].div(summ);
				}
			}

			return new BatteryBankDistribution(distribution, summ);
		}

		private Fraction distributeEnergy(boolean mode, Fraction howMuch, boolean simulate) {
			BatteryBankDistribution distribution = getDistribution(mode);

			if (distribution.max_throughput.compareTo(Fraction.ZERO) == 0)
				return Fraction.ZERO;

			Fraction[] distList = distribution.distribution;
			Fraction summ = Fraction.ZERO;

			for (int i = 0; i < tile.battery_container.getContainerSize(); i++) {
				if (!distList[i].equals(Fraction.ZERO)) {
					ItemStack stack = tile.battery_container.getItem(i);

					if (!stack.isEmpty()) {
						Optional<IMatteryEnergyStorage> cap = stack.getCapability(MatteryCapability.ENERGY).resolve();

						if (cap.isPresent()) {
							Fraction diff = mode ? cap.get().receiveEnergyOuter(howMuch.times(distList[i]), simulate) : cap.get().extractEnergyOuter(howMuch.times(distList[i]), simulate);
							summ = summ.plus(diff);
						} else {
							Optional<IEnergyStorage> cap2 = stack.getCapability(CapabilityEnergy.ENERGY).resolve();

							if (cap2.isPresent()) {
								Fraction diff = mode ? MatteryCapability.floodFE(cap2.get(), howMuch.times(distList[i]), simulate) : MatteryCapability.drainFE(cap2.get(), howMuch.times(distList[i]), simulate);
								summ = summ.plus(diff);
							}
						}
					}
				}
			}

			return summ;
		}

		@Nonnull
		@Override
		public Fraction extractEnergyInner(Fraction howMuch, boolean simulate) {
			return distributeEnergy(false, howMuch, simulate);
		}

		@Nonnull
		@Override
		public Fraction receiveEnergyOuter(Fraction howMuch, boolean simulate) {
			if (mode == BankMode.EXTRACT)
				return Fraction.ZERO;

			return receiveEnergyInner(howMuch, simulate);
		}

		@Nonnull
		@Override
		public Fraction receiveEnergyInner(Fraction howMuch, boolean simulate) {
			return distributeEnergy(true, howMuch, simulate);
		}

		@Nonnull
		@Override
		public Fraction getBatteryLevel() {
			Fraction result = Fraction.ZERO;

			for (int i = 0; i < tile.battery_container.getContainerSize(); i++) {
				ItemStack stack = tile.battery_container.getItem(i);

				if (!stack.isEmpty()) {
					Optional<IMatteryEnergyStorage> cap = stack.getCapability(MatteryCapability.ENERGY).resolve();

					if (cap.isPresent()) {
						result = result.plus(cap.get().getBatteryLevel());
					} else {
						Optional<IEnergyStorage> cap2 = stack.getCapability(CapabilityEnergy.ENERGY).resolve();

						if (cap2.isPresent()) {
							result = result.plus(cap2.get().getEnergyStored());
						}
					}
				}
			}

			return result;
		}

		@Nonnull
		@Override
		public Fraction getMaxBatteryLevel() {
			Fraction result = Fraction.ZERO;

			for (int i = 0; i < tile.battery_container.getContainerSize(); i++) {
				ItemStack stack = tile.battery_container.getItem(i);

				if (!stack.isEmpty()) {
					Optional<IMatteryEnergyStorage> cap = stack.getCapability(MatteryCapability.ENERGY).resolve();

					if (cap.isPresent()) {
						result = result.plus(cap.get().getMaxBatteryLevel());
					} else {
						Optional<IEnergyStorage> cap2 = stack.getCapability(CapabilityEnergy.ENERGY).resolve();

						if (cap2.isPresent()) {
							result = result.plus(cap2.get().getMaxEnergyStored());
						}
					}
				}
			}

			return result;
		}
	}

	public final BatteryBankEnergy energy_receiver = new BatteryBankEnergy(this, BatteryBankEnergy.BankMode.RECEIVE);
	private LazyOptional<BatteryBankEnergy> energy_receiver_resolver = LazyOptional.of(() -> energy_receiver);

	public final BatteryBankEnergy energy_extractor = new BatteryBankEnergy(this, BatteryBankEnergy.BankMode.EXTRACT);
	private LazyOptional<BatteryBankEnergy> energy_extractor_resolver = LazyOptional.of(() -> energy_extractor);

	public final BatteryBankEnergy energy = new BatteryBankEnergy(this, BatteryBankEnergy.BankMode.BIDIRECTIONAL);
	private LazyOptional<BatteryBankEnergy> energy_resolver = LazyOptional.of(() -> energy);

	protected boolean valid = true;

	public BlockEntityBatteryBank(BlockPos p_155229_, BlockState p_155230_) {
		super(Registry.BlockEntities.BATTERY_BANK, p_155229_, p_155230_);
	}

	@Override
	public void saveAdditional(CompoundTag nbt) {
		super.saveAdditional(nbt);
		nbt.put("battery_bank", battery_container.serializeNBT());
	}

	@Override
	public void load(CompoundTag nbt) {
		battery_container.deserializeNBT(nbt.get("battery_bank"));
		super.load(nbt);
	}

	private final TranslatableComponent MACHINE_NAME = new TranslatableComponent("block.overdrive_that_matters.battery_bank");

	@Override
	protected Component getDefaultDisplayName() {
		return MACHINE_NAME;
	}

	@Nullable
	@Override
	public AbstractContainerMenu createMenu(int containerID, Inventory inventory, Player ply) {
		return new BatteryBankMenu(containerID, inventory, this, this.battery_container);
	}

	@Override
	public void invalidateCaps() {
		super.invalidateCaps();
		valid = false;
		energy_receiver_resolver.invalidate();
		energy_extractor_resolver.invalidate();
		energy_resolver.invalidate();
	}

	@Override
	public void reviveCaps() {
		super.reviveCaps();
		valid = true;
		energy_receiver_resolver = LazyOptional.of(() -> energy_receiver);
		energy_extractor_resolver = LazyOptional.of(() -> energy_extractor);
		energy_resolver = LazyOptional.of(() -> energy);
	}

	@Override
	public void setLevel(Level p_155231_) {
		super.setLevel(p_155231_);
		tickOnceServer(this::checkSurroundings);
	}

	@Nonnull
	public <T> LazyOptional<T> getCapability(@Nonnull Capability<T> cap, @Nullable Direction side) {
		if (valid && (cap == MatteryCapability.ENERGY || cap == CapabilityEnergy.ENERGY)) {
			if (side == null)
				return energy_resolver.cast();

			if (side == getBlockState().getValue(BlockMatteryRotatable.FACING))
				return energy_extractor_resolver.cast();

			return energy_receiver_resolver.cast();
		}

		if (valid && cap == CapabilityItemHandler.ITEM_HANDLER_CAPABILITY)
			return item_handler_resolver.cast();

		return super.getCapability(cap, side);
	}

	protected LazyOptional<IMatteryEnergyStorage> output_cap_mte = LazyOptional.empty();
	protected LazyOptional<IEnergyStorage> output_cap_fe = LazyOptional.empty();

	protected void _checkSurroundings() {
		if (level instanceof ServerLevel level)
			checkSurroundings(level);
	}

	public void checkSurroundings(Level level) {
		if (isRemoved())
			return;

		final var get_entity = level.getBlockEntity(getBlockPos().offset(getBlockState().getValue(BlockMatteryRotatable.FACING).getNormal()));

		if (get_entity == null) {
			output_cap_mte = LazyOptional.empty();
			output_cap_fe = LazyOptional.empty();
			return;
		}

		output_cap_mte = getAndBind(output_cap_mte, get_entity, MatteryCapability.ENERGY, getBlockState().getValue(BlockMatteryRotatable.FACING).getOpposite(), this::_checkSurroundings);
		output_cap_fe = getAndBind(output_cap_fe, get_entity, CapabilityEnergy.ENERGY, getBlockState().getValue(BlockMatteryRotatable.FACING).getOpposite(), this::_checkSurroundings);
	}

	public static <T extends BlockEntity> void tick(Level level, BlockPos blockPos, BlockState blockState, T t) {
		if (t instanceof BlockEntityBatteryBank tile) {
			if (tile.isBlockedByRedstone())
				return;

			if (tile.output_cap_mte.isPresent()) {
				final var cap = tile.output_cap_mte.resolve().get();

				if (cap.canReceive()) {
					BatteryBankDistribution distribution = tile.energy.getDistribution(false);
					Fraction diff = cap.receiveEnergyOuter(distribution.max_throughput, true);
					diff = tile.energy.extractEnergyOuter(diff, false);
					cap.receiveEnergyOuter(diff, false);
				}
			} else if (tile.output_cap_fe.isPresent()) {
				final var cap = tile.output_cap_fe.resolve().get();

				if (cap.canReceive()) {
					BatteryBankDistribution distribution = tile.energy.getDistribution(false);
					Fraction diff = MatteryCapability.floodFE(cap, distribution.max_throughput, true);
					diff = tile.energy.extractEnergyOuter(diff, false);
					MatteryCapability.floodFE(cap, diff, false);
				}
			}
		}
	}
}
