package ru.dbotthepony.mc.otm.block.entity;

import net.minecraft.MethodsReturnNonnullByDefault;
import net.minecraft.core.BlockPos;
import net.minecraft.core.Direction;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.network.chat.Component;
import net.minecraft.network.chat.TranslatableComponent;
import net.minecraft.world.entity.player.Inventory;
import net.minecraft.world.entity.player.Player;
import net.minecraft.world.inventory.AbstractContainerMenu;
import net.minecraft.world.item.ItemStack;
import net.minecraft.world.level.Level;
import net.minecraft.world.level.block.Block;
import net.minecraft.world.level.block.state.BlockState;
import net.minecraftforge.common.capabilities.Capability;
import net.minecraftforge.common.util.LazyOptional;
import ru.dbotthepony.mc.otm.block.BlockBatteryBank;
import ru.dbotthepony.mc.otm.capability.matter.IMatterGridCell;
import ru.dbotthepony.mc.otm.core.Fraction;
import ru.dbotthepony.mc.otm.matter.MatterGrid;
import ru.dbotthepony.mc.otm.Registry;
import ru.dbotthepony.mc.otm.capability.matter.IMatterHandler;
import ru.dbotthepony.mc.otm.capability.matter.IPatternStorage;
import ru.dbotthepony.mc.otm.capability.MatteryCapability;
import ru.dbotthepony.mc.otm.container.MatteryContainer;
import ru.dbotthepony.mc.otm.menu.MatterCapacitorBankMenu;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.ParametersAreNonnullByDefault;
import java.math.BigDecimal;
import java.util.Optional;

@MethodsReturnNonnullByDefault
@ParametersAreNonnullByDefault
public class BlockEntityMatterCapacitorBank extends BlockEntityMattery implements IMatterGridCell {
	public final IMatterHandler matter = new IMatterHandler() {
		@Nonnull
		@Override
		public Fraction getStoredMatter() {
			Fraction summ = Fraction.ZERO;

			for (int i = 0; i < matter_container.getContainerSize(); i++) {
				ItemStack stack = matter_container.getItem(i);

				if (!stack.isEmpty()) {
					Optional<IMatterHandler> handler = stack.getCapability(MatteryCapability.MATTER).resolve();

					if (handler.isPresent()) {
						summ = summ.plus(handler.get().getStoredMatter());
					}
				}
			}

			return summ;
		}

		@Nonnull
		@Override
		public Fraction getMaxStoredMatter() {
			Fraction summ = Fraction.ZERO;

			for (int i = 0; i < matter_container.getContainerSize(); i++) {
				ItemStack stack = matter_container.getItem(i);

				if (!stack.isEmpty()) {
					Optional<IMatterHandler> handler = stack.getCapability(MatteryCapability.MATTER).resolve();

					if (handler.isPresent()) {
						summ = summ.plus(handler.get().getMaxStoredMatter());
					}
				}
			}

			return summ;
		}

		@Nonnull
		@Override
		public Fraction receiveMatterOuter(Fraction howMuch, boolean simulate) {
			return receiveMatterInner(howMuch, simulate);
		}

		@Nonnull
		@Override
		public Fraction receiveMatterInner(Fraction howMuch, boolean simulate) {
			Fraction summ = Fraction.ZERO;

			for (int i = 0; i < matter_container.getContainerSize(); i++) {
				ItemStack stack = matter_container.getItem(i);

				if (!stack.isEmpty()) {
					Optional<IMatterHandler> handler = stack.getCapability(MatteryCapability.MATTER).resolve();

					if (handler.isPresent()) {
						var diff = handler.get().receiveMatterOuter(howMuch, simulate);
						summ = summ.plus(diff);
						howMuch = howMuch.minus(diff);

						if (howMuch.compareTo(Fraction.ZERO) == 0) {
							break;
						}
					}
				}
			}

			return summ;
		}

		@Nonnull
		@Override
		public Fraction extractMatterOuter(Fraction howMuch, boolean simulate) {
			return extractMatterInner(howMuch, simulate);
		}

		@Nonnull
		@Override
		public Fraction extractMatterInner(Fraction howMuch, boolean simulate) {
			Fraction summ = Fraction.ZERO;

			for (int i = 0; i < matter_container.getContainerSize(); i++) {
				ItemStack stack = matter_container.getItem(i);

				if (!stack.isEmpty()) {
					Optional<IMatterHandler> handler = stack.getCapability(MatteryCapability.MATTER).resolve();

					if (handler.isPresent()) {
						var diff = handler.get().extractMatterOuter(howMuch, simulate);
						summ = summ.plus(diff);
						howMuch = howMuch.minus(diff);

						if (howMuch.compareTo(Fraction.ZERO) == 0) {
							break;
						}
					}
				}
			}

			return summ;
		}

		@Nonnull
		@Override
		public MatterDirection getDirection() {
			return MatterDirection.BIDIRECTIONAL;
		}
	};

	private LazyOptional<IMatterHandler> resolver = LazyOptional.of(() -> matter);

	public final MatteryContainer matter_container = new MatteryContainer(this::setChanged, 6 * 2) {
		@Override
		public void setChanged(int slot, ItemStack new_state, ItemStack old_state) {
			super.setChanged(slot, new_state, old_state);

			if (level != null) {
				var state = getBlockState();

				for (int i = 0; i < BlockBatteryBank.BATTERY_SLOTS_PROPS.length; i++) {
					state = state.setValue(BlockBatteryBank.BATTERY_SLOTS_PROPS[i], getItem(i).getCapability(MatteryCapability.MATTER).isPresent());
				}

				if (state != getBlockState()) {
					level.setBlock(getBlockPos(), state, Block.UPDATE_CLIENTS);
				}
			}
		}
	};

	public BlockEntityMatterCapacitorBank(BlockPos p_155229_, BlockState p_155230_) {
		super(Registry.BlockEntities.MATTER_CAPACITOR_BANK, p_155229_, p_155230_);
	}

	private final TranslatableComponent MACHINE_NAME = new TranslatableComponent("block.overdrive_that_matters.matter_capacitor_bank");

	@Override
	protected Component getDefaultDisplayName() {
		return MACHINE_NAME;
	}

	@Override
	public void saveAdditional(CompoundTag nbt) {
		super.saveAdditional(nbt);
		nbt.put("matter_container", matter_container.serializeNBT());
	}

	@Override
	public void load(CompoundTag nbt) {
		matter_container.deserializeNBT(nbt.get("matter_container"));
		super.load(nbt);
	}

	@Nullable
	@Override
	public AbstractContainerMenu createMenu(int containerID, Inventory inventory, Player ply) {
		return new MatterCapacitorBankMenu(containerID, inventory, this);
	}

	private boolean valid = true;

	@Override
	public void invalidateCaps() {
		super.invalidateCaps();
		valid = false;
		resolver.invalidate();
		resolver_grid.invalidate();
	}

	@Override
	public void reviveCaps() {
		super.reviveCaps();
		valid = true;
		resolver = LazyOptional.of(() -> matter);
		resolver_grid = LazyOptional.of(() -> this);
	}

	private LazyOptional<IMatterGridCell> resolver_grid = LazyOptional.of(() -> this);

	@Nonnull
	@Override
	public <T> LazyOptional<T> getCapability(@Nonnull Capability<T> cap, @Nullable Direction side) {
		if (valid) {
			if (cap == MatteryCapability.MATTER)
				return resolver.cast();

			if (cap == MatteryCapability.MATTER_CELL)
				return resolver_grid.cast();
		}

		return super.getCapability(cap, side);
	}

	private MatterGrid grid;

	@Nullable
	@Override
	public MatterGrid getMatterGrid() {
		return grid;
	}

	@Override
	public void setRemoved() {
		super.setRemoved();

		if (grid != null)
			grid.remove(this);
	}

	@Override
	public void setLevel(Level p_155231_) {
		super.setLevel(p_155231_);

		if (!p_155231_.isClientSide)
			MatterGrid.scheduleDiscoverNeighbours(this, getBlockPos(), p_155231_);
	}

	@Nullable
	@Override
	public IMatterHandler getMatterHandler() {
		return valid ? matter : null;
	}

	@Nullable
	@Override
	public IPatternStorage getPatternStorage() {
		return null;
	}

	@Override
	public boolean isValidMatterCell() {
		return valid;
	}

	@Override
	public void setMatterGrid(MatterGrid grid) {
		this.grid = grid;
	}
}
