package ru.dbotthepony.mc.otm.capability.android;

import net.minecraft.ChatFormatting;
import net.minecraft.MethodsReturnNonnullByDefault;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.nbt.ListTag;
import net.minecraft.nbt.Tag;
import net.minecraft.network.chat.TranslatableComponent;
import net.minecraft.resources.ResourceLocation;
import net.minecraft.server.level.ServerPlayer;
import net.minecraft.world.entity.Entity;
import net.minecraft.world.entity.player.Player;
import net.minecraftforge.common.util.LazyOptional;
import net.minecraftforge.event.AttachCapabilitiesEvent;
import net.minecraftforge.event.entity.player.PlayerEvent;
import net.minecraftforge.eventbus.api.SubscribeEvent;
import ru.dbotthepony.mc.otm.Registry;
import ru.dbotthepony.mc.otm.android.AndroidResearch;
import ru.dbotthepony.mc.otm.android.AndroidResearchType;
import ru.dbotthepony.mc.otm.capability.MatteryCapability;
import ru.dbotthepony.mc.otm.core.Fraction;
import ru.dbotthepony.mc.otm.network.android.AndroidResearchPacket;
import ru.dbotthepony.mc.otm.network.android.AndroidStatusPacket;

import javax.annotation.Nonnull;
import javax.annotation.ParametersAreNonnullByDefault;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Objects;

@MethodsReturnNonnullByDefault
@ParametersAreNonnullByDefault
public class AndroidCapabilityPlayer extends AndroidCapability {
	public boolean is_android = false;
	private boolean network_is_android = false;

	public boolean will_become_android = false;
	private boolean network_will_become_android = false;

	private final ArrayList<AndroidResearch> research_list = new ArrayList<>();

	@Override
	public void invalidateNetworkState() {
		super.invalidateNetworkState();
		network_is_android = false;
		network_will_become_android = false;

		for (var instance : research_list) {
			instance.dirty = true;
		}
	}

	@Override
	public boolean isAndroid() {
		return is_android;
	}

	public boolean isEverAndroid() {
		return is_android || will_become_android;
	}

	public void becomeAndroidSoft() {
		if (is_android || will_become_android)
			return;

		will_become_android = true;

		if (ent instanceof ServerPlayer ply) {
			ply.displayClientMessage(new TranslatableComponent("otm.pill.message").withStyle(ChatFormatting.GRAY), false);
		}
	}

	public void becomeAndroid() {
		if (is_android)
			return;

		is_android = true;
		will_become_android = false;
		energy_stored = new Fraction(60_000);
		energy_stored_max = new Fraction(60_000);
	}

	public void becomeAndroidAndKill() {
		if (is_android)
			return;

		becomeAndroid();
		ply.hurt(Registry.DAMAGE_BECOME_ANDROID, ply.getMaxHealth() * 2);
	}

	public void becomeHumane() {
		if (will_become_android)
			will_become_android = false;

		if (!is_android)
			return;

		is_android = false;

		energy_stored = new Fraction(0);
		energy_stored_max = new Fraction(60_000);

		dropBattery();
	}

	public void becomeHumaneAndKill() {
		if (will_become_android)
			will_become_android = false;

		if (!is_android)
			return;

		becomeHumane();
		ply.hurt(Registry.DAMAGE_BECOME_HUMANE, ply.getMaxHealth() * 2);
	}

	@Override
	public void deserializeNBT(CompoundTag compound) {
		super.deserializeNBT(compound);

		is_android = compound.getBoolean("is_android");
		will_become_android = compound.getBoolean("will_become_android");

		research_list.clear();

		var list = compound.getList("research", Tag.TAG_COMPOUND);

		for (var _tag : list) {
			if (_tag instanceof CompoundTag tag) {
				var research = Registry.ANDROID_RESEARCH.getValue(new ResourceLocation(tag.getString("id")));

				if (research != null) {
					var instance = research.factory(this);
					instance.deserializeNBT(tag);
					research_list.add(instance);
				}
			}
		}
	}

	@Nonnull
	@Override
	public CompoundTag serializeNBT() {
		CompoundTag tag = super.serializeNBT();
		tag.putBoolean("is_android", is_android);
		tag.putBoolean("will_become_android", will_become_android);

		var list = new ListTag();

		for (var instance : research_list) {
			var _tag = new CompoundTag();
			instance.serializeNBT(_tag);
			_tag.putString("id", Objects.requireNonNull(instance.type.getRegistryName()).toString());
			list.add(_tag);
		}

		tag.put("research", list);

		return tag;
	}

	@SuppressWarnings("unchecked")
	public <T extends AndroidResearch> T getResearch(AndroidResearchType<T> type) {
		for (var instance : research_list) {
			if (instance.type == type) {
				return (T) instance;
			}
		}

		var instance = type.factory(this);
		research_list.add(instance);
		return instance;
	}

	@SubscribeEvent
	@SuppressWarnings("unused")
	public static void onAttachCapabilityEvent(AttachCapabilitiesEvent<Entity> event) {
		if (event.getObject() instanceof Player ply) {
			event.addCapability(Registry.Names.ANDROID_CAPABILITY, new AndroidCapabilityPlayer(ply));
		}
	}

	@SubscribeEvent
	@SuppressWarnings("unused")
	public static void onPlayerChangeDimensionEvent(PlayerEvent.PlayerChangedDimensionEvent event) {
		event.getPlayer().getCapability(MatteryCapability.ANDROID).ifPresent(IAndroidCapability::invalidateNetworkState);
	}

	@SubscribeEvent
	@SuppressWarnings("unused")
	public static void onPlayerCloneEvent(PlayerEvent.Clone event) {
		event.getPlayer().getCapability(MatteryCapability.ANDROID).ifPresent((cap) -> {
			LazyOptional<IAndroidCapability> resolver = event.getOriginal().getCapability(MatteryCapability.ANDROID);

			if (!resolver.isPresent() || resolver.resolve().isEmpty()) {
				event.getOriginal().reviveCaps();
				resolver = event.getOriginal().getCapability(MatteryCapability.ANDROID);
			}

			if (!resolver.isPresent() || resolver.resolve().isEmpty()) {
				event.getOriginal().invalidateCaps();
				return;
			}

			var original = (AndroidCapabilityPlayer) resolver.resolve().get();

			if (original.will_become_android && event.isWasDeath()) {
				original.becomeAndroid();

				if (event.getPlayer() instanceof ServerPlayer ply) {
					ply.displayClientMessage(new TranslatableComponent("otm.pill.message_finish").withStyle(ChatFormatting.DARK_RED), false);
				}
			}

			cap.deserializeNBT(original.serializeNBT());
			cap.invalidateNetworkState();
			event.getOriginal().invalidateCaps();
		});
	}

	public int last_jump_ticks = 14;
	public final Player ply;

	public AndroidCapabilityPlayer(Player ent) {
		super(ent);
		ply = ent;
	}

	@Override
	protected void tickNetwork() {
		super.tickNetwork();

		if (is_android != network_is_android) {
			network_is_android = is_android;
			sendNetwork(new AndroidStatusPacket(AndroidStatusPacket.Type.IS_ANDROID, is_android));
		}

		if (will_become_android != network_will_become_android) {
			network_will_become_android = will_become_android;
			sendNetwork(new AndroidStatusPacket(AndroidStatusPacket.Type.WILL_BECOME_ANDROID, will_become_android));
		}

		for (var instance : research_list) {
			if (instance.dirty) {
				instance.dirty = false;
				sendNetwork(new AndroidResearchPacket(instance));
			}
		}
	}

	public static final Fraction ENERGY_FOR_HUNGER_POINT = new Fraction(1000);

	public int sleep_ticks = 0;
	public static final int SLEEP_TICKS_LIMIT = 80;

	@Override
	protected void tickInnerClientAlways() {
		super.tickInnerClientAlways();

		if (will_become_android) {
			if (ent.isSleeping()) {
				sleep_ticks++;
			} else {
				sleep_ticks = 0;
			}
		}
	}

	@Override
	protected void tickServerAlways() {
		super.tickServerAlways();

		if (will_become_android) {
			if (ent.isSleeping()) {
				sleep_ticks++;

				if (sleep_ticks > SLEEP_TICKS_LIMIT) {
					becomeAndroid();
					sleep_ticks = 0;

					if (ent instanceof ServerPlayer ply) {
						ply.displayClientMessage(new TranslatableComponent("otm.pill.message_finish").withStyle(ChatFormatting.DARK_RED), false);
					}
				}
			} else {
				sleep_ticks = 0;
			}
		}
	}

	@Override
	public void tickServer() {
		super.tickServer();

		// TODO: Maybe passive drain?
		// extractEnergyInner(BigDecimal.valueOf(new Random().nextDouble()), false);

		if (ply.isSwimming() && !hasFeature(Registry.AndroidFeatures.AIR_BAGS)) {
			ply.setSwimming(false);
		}

		var stats = ply.getFoodData();

		while (stats.getFoodLevel() < 18 && this.extractEnergyInner(ENERGY_FOR_HUNGER_POINT, true).compareTo(ENERGY_FOR_HUNGER_POINT) == 0) {
			this.extractEnergyInner(ENERGY_FOR_HUNGER_POINT, false);
			stats.setFoodLevel(stats.getFoodLevel() + 1);
		}

		// "block" quick regeneration
		// also cause power to generate while in peaceful
		while (stats.getFoodLevel() > 18 && this.receiveEnergyInner(ENERGY_FOR_HUNGER_POINT, true).compareTo(ENERGY_FOR_HUNGER_POINT) == 0) {
			this.receiveEnergyInner(ENERGY_FOR_HUNGER_POINT, false);
			stats.setFoodLevel(stats.getFoodLevel() - 1);
		}

		var food_level = (float) stats.getFoodLevel();

		if (stats.getSaturationLevel() < food_level) {
			Fraction extracted = this.extractEnergyInner(ENERGY_FOR_HUNGER_POINT.times(food_level - stats.getSaturationLevel()), false);
			stats.setSaturation(stats.getSaturationLevel() + extracted.div(ENERGY_FOR_HUNGER_POINT).toFloat());
		}

		if (stats.getExhaustionLevel() > 0f) {
			Fraction extracted = this.extractEnergyInner(ENERGY_FOR_HUNGER_POINT.times(stats.getExhaustionLevel() / 4f), false);
			stats.setExhaustion(stats.getExhaustionLevel() - extracted.div(ENERGY_FOR_HUNGER_POINT).toFloat() * 4f);
		}
	}
}
