package ru.dbotthepony.mc.otm.menu;

import net.minecraft.network.FriendlyByteBuf;
import net.minecraft.world.SimpleContainer;
import net.minecraft.world.entity.player.Inventory;
import net.minecraft.world.entity.player.Player;
import net.minecraft.world.item.ItemStack;
import net.minecraftforge.energy.CapabilityEnergy;
import net.minecraftforge.network.NetworkEvent;
import ru.dbotthepony.mc.otm.Registry;
import ru.dbotthepony.mc.otm.block.entity.BlockEntityDriveViewer;
import ru.dbotthepony.mc.otm.capability.MatteryCapability;
import ru.dbotthepony.mc.otm.capability.drive.IMatteryDrive;
import ru.dbotthepony.mc.otm.item.ItemPortableCondensationDrive;
import ru.dbotthepony.mc.otm.menu.data.INetworkedItemViewSupplier;
import ru.dbotthepony.mc.otm.menu.data.NetworkedItemView;
import ru.dbotthepony.mc.otm.menu.slot.MatterySlot;
import ru.dbotthepony.mc.otm.storage.ItemStackWrapper;

import javax.annotation.Nullable;
import java.util.function.Supplier;

public class DriveViewerMenu extends PoweredMatteryMenu implements INetworkedItemViewSupplier {
	public final NetworkedItemView view;
	public MatterySlot drive_slot;
	public IMatteryDrive<ItemStackWrapper> last_drive;

	@Override
	public NetworkedItemView getNetworkedItemView() {
		return view;
	}

	public DriveViewerMenu(int containerID, Inventory inventory) {
		this(containerID, inventory, null);
	}

	public DriveViewerMenu(int containerID, Inventory inventory, @Nullable BlockEntityDriveViewer tile) {
		super(Registry.Menus.DRIVE_VIEWER, containerID, inventory, tile);

		var container = tile != null ? tile.drive_slot : new SimpleContainer(1);

		drive_slot = new MatterySlot(container, 0) {
			@Override
			public boolean mayPlace(ItemStack stack) {
				return stack.getCapability(MatteryCapability.DRIVE).isPresent();
			}
		};

		view = new NetworkedItemView(inventory.player, this, tile == null) {
			@Override
			public int calculateIOAmount(int desired_amount) {
				if (tile != null)
					return tile.getIOItemCount(desired_amount, true);

				return super.calculateIOAmount(desired_amount);
			}

			@Override
			public void doneIOAmount(int done_amount) {
				if (tile != null) {
					tile.getIOItemCount(done_amount, false);
					return;
				}

				super.doneIOAmount(done_amount);
			}
		};

		addSlot(drive_slot);
		addBatterySlot();
		addInventorySlots();
	}

	@Override
	@SuppressWarnings("unchecked")
	public void broadcastChanges() {
		super.broadcastChanges();

		if (tile != null) {
			var get = ((BlockEntityDriveViewer) tile).drive_slot.getItem(0);

			if (get.isEmpty()) {
				last_drive = null;
			} else {
				var get_cap = get.getCapability(MatteryCapability.DRIVE).resolve();

				if (get_cap.isEmpty() || get_cap.get().storageIdentity() != ItemStackWrapper.class) {
					last_drive = null;
				} else {
					last_drive = (IMatteryDrive<ItemStackWrapper>) get_cap.get();
				}
			}

			view.setComponent(last_drive);
			view.network();
		}
	}

	@Override
	protected int getWorkingSlotStart() {
		return 0;
	}

	@Override
	protected int getWorkingSlotEnd() {
		return 2;
	}

	@Override
	public void removed(Player p_38940_) {
		super.removed(p_38940_);

		if (last_drive != null) {
			last_drive.removeListener(view);
			view.clear();
		}
	}

	@Override
	public ItemStack quickMoveStack(Player ply, int slot_index) {
		var slot = slots.get(slot_index);
		var item = slot.getItem();

		if (item.isEmpty() || item.getCapability(MatteryCapability.DRIVE).isPresent() || item.getCapability(CapabilityEnergy.ENERGY).isPresent())
			return super.quickMoveStack(ply, slot_index);

		if (last_drive == null || tile == null)
			return ItemStack.EMPTY;

		var tile = (BlockEntityDriveViewer) this.tile;
		int amount = tile.getIOItemCount(item.getCount(), true);

		if (amount == 0)
			return ItemStack.EMPTY;

		if (amount == item.getCount()) {
			var remaining = last_drive.insertStack(new ItemStackWrapper(item), false);

			if (remaining.getCount().toInt() == item.getCount())
				return ItemStack.EMPTY;

			if (remaining.isEmpty()) {
				var copy = item.copy();
				slot.set(ItemStack.EMPTY);
				tile.getIOItemCount(item.getCount(), false);
				return copy;
			}

			var copy = item.copy();
			tile.getIOItemCount(item.getCount() - remaining.getCount().toInt(), false);
			item.shrink(remaining.getCount().toInt());
			slot.setChanged();

			return copy;
		}

		var copy_insert = item.copy();
		copy_insert.setCount(amount);

		var remaining = last_drive.insertStack(new ItemStackWrapper(copy_insert), false);

		if (remaining.getCount().toInt() == copy_insert.getCount())
			return ItemStack.EMPTY;

		var copy = item.copy();
		tile.getIOItemCount(amount - remaining.getCount().toInt(), false);
		item.shrink(amount - remaining.getCount().toInt());
		slot.setChanged();

		return copy;
	}

	public enum FilterSwitch {
		MATCH_NBT,
		MATCH_TAG,
		BLACKLIST
	}

	public ItemPortableCondensationDrive.FilterSettings getFilter() {
		if (drive_slot.getItem().isEmpty())
			return null;

		if (drive_slot.getItem().getItem() instanceof ItemPortableCondensationDrive item)
			return item.getFilterSettings(drive_slot.getItem());

		return null;
	}

	public record FilterSetPacket(int id, int slot, ItemStack value) {
		public void write(FriendlyByteBuf buffer) {
			buffer.writeInt(id);
			buffer.writeInt(slot);
			buffer.writeItem(value);
		}

		public static FilterSetPacket read(FriendlyByteBuf buffer) {
			return new FilterSetPacket(buffer.readInt(), buffer.readInt(), buffer.readItem());
		}

		public void play(Supplier<NetworkEvent.Context> context) {
			context.get().setPacketHandled(true);

			if (slot < 0 || slot >= ItemPortableCondensationDrive.FilterSettings.MAX_FILTERS)
				return;

			if (value.getCount() > 1)
				value.setCount(1);

			context.get().enqueueWork(() -> {
				var ply = context.get().getSender();

				if (ply.containerMenu instanceof DriveViewerMenu menu && menu.containerId == id) {
					if (menu.drive_slot.getItem().isEmpty())
						return;

					if (menu.drive_slot.getItem().getItem() instanceof ItemPortableCondensationDrive drive) {
						var filter = drive.getFilterSettings(menu.drive_slot.getItem());
						filter.items[slot] = value;
						filter.serializeNBT(menu.drive_slot.getItem());
					}
				}
			});
		}
	}

	public record FilterSwitchPacket(int id, FilterSwitch type, boolean value) {
		public void write(FriendlyByteBuf buffer) {
			buffer.writeInt(id);
			buffer.writeEnum(type);
			buffer.writeBoolean(value);
		}

		public static FilterSwitchPacket read(FriendlyByteBuf buffer) {
			return new FilterSwitchPacket(buffer.readInt(), buffer.readEnum(FilterSwitch.class), buffer.readBoolean());
		}

		public void play(Supplier<NetworkEvent.Context> context) {
			context.get().setPacketHandled(true);
			context.get().enqueueWork(() -> {
				var ply = context.get().getSender();

				if (ply.containerMenu instanceof DriveViewerMenu menu && menu.containerId == id) {
					var settings = menu.getFilter();

					if (settings == null)
						return;

					switch (type) {
						case MATCH_NBT -> {
							settings.match_nbt = value;
						}

						case MATCH_TAG -> {
							settings.match_tag = value;
						}

						case BLACKLIST -> {
							settings.blacklist = value;
						}
					}

					settings.serializeNBT(menu.drive_slot.getItem());
				}
			});
		}
	}
}
