package ru.dbotthepony.mc.otm.network;

import net.minecraft.network.FriendlyByteBuf;
import net.minecraft.world.item.Item;
import net.minecraftforge.api.distmarker.Dist;
import net.minecraftforge.fml.DistExecutor;
import net.minecraftforge.network.NetworkEvent;
import net.minecraftforge.registries.ForgeRegistry;
import net.minecraftforge.registries.RegistryManager;
import ru.dbotthepony.mc.otm.core.Fraction;
import ru.dbotthepony.mc.otm.matter.MatterRegistry;

import java.math.BigDecimal;
import java.util.HashMap;
import java.util.Map;
import java.util.function.Supplier;

public record MatterRegistryPacket(Map<Item, Fraction> list) {
	public void write(FriendlyByteBuf buffer) {
		buffer.writeInt(list.size());
		var registry = (ForgeRegistry<Item>) RegistryManager.ACTIVE.getRegistry(Item.class);

		for (var entry : list.entrySet()) {
			buffer.writeInt(registry.getID(entry.getKey()));
			entry.getValue().write(buffer);
		}
	}

	public static MatterRegistryPacket read(FriendlyByteBuf buffer) {
		var size = buffer.readInt();
		var registry = (ForgeRegistry<Item>) RegistryManager.ACTIVE.getRegistry(Item.class);

		var map = new HashMap<Item, Fraction>();

		for (int i = 0; i < size; i++) {
			map.put(registry.getValue(buffer.readInt()), Fraction.read(buffer));
		}

		return new MatterRegistryPacket(map);
	}

	public void playClient() {
		MatterRegistry.ITEMS.clear();
		MatterRegistry.ITEMS.putAll(list);
	}

	public void play(Supplier<NetworkEvent.Context> context) {
		context.get().setPacketHandled(true);

		DistExecutor.unsafeRunWhenOn(Dist.CLIENT, () -> this::playClient);
	}
}
