package ru.dbotthepony.mc.otm.tests

import org.junit.jupiter.api.DisplayName
import org.junit.jupiter.api.Test
import ru.dbotthepony.mc.otm.core.Fraction
import java.math.BigDecimal
import java.math.BigInteger

object FractionTests {
	@Test
	@DisplayName("Fraction declaration")
	fun declaration() {
		println("BigInteger 1/1 == ${Fraction(BigInteger.valueOf(1), BigInteger.valueOf(1))}")
		println("BigDecimal 1 == ${Fraction(BigDecimal("1"))}")

		var one = BigDecimal("1.00")

		println("Unscaled ${one.unscaledValue()} with scale ${one.scale()} of $one")
		println("BigDecimal $one == ${Fraction(one)}")
		assert(Fraction(one).compareTo(Fraction.ONE) == 0)

		one = BigDecimal("1.0000")

		println("Unscaled ${one.unscaledValue()} with scale ${one.scale()} of $one")
		println("BigDecimal $one == ${Fraction(one)}")
		assert(Fraction(one).compareTo(Fraction.ONE) == 0)

		println("1/2 == ${Fraction(1, 2)}")
		println("-1/2 == ${Fraction(-1, 2)}")
		println("1/-2 == ${Fraction(1, -2)}")
		println("-1/-2 == ${Fraction(-1, -2)}")

		println("canonical 1/2 == ${Fraction(1, 2).canonize()}")
		println("canonical -1/2 == ${Fraction(-1, 2).canonize()}")
		println("canonical 1/-2 == ${Fraction(1, -2).canonize()}")
		println("canonical -1/-2 == ${Fraction(-1, -2).canonize()}")
	}

	@Test
	@DisplayName("Fraction comparison")
	fun equality() {
		assert(Fraction(1).compareTo(Fraction.ONE) == 0)
		assert(Fraction(1.0).compareTo(Fraction.ONE) == 0)
		assert(Fraction(1.0F).compareTo(Fraction.ONE) == 0)
		assert(Fraction(1, 1).compareTo(Fraction.ONE) == 0)
		assert(Fraction(1.0, 1.0).compareTo(Fraction.ONE) == 0)
		assert(Fraction(1.0F, 1.0F).compareTo(Fraction.ONE) == 0)

		assert(Fraction(1, 2).compareTo(Fraction.HALF) == 0)
		assert(Fraction(1.0, 2.0).compareTo(Fraction.HALF) == 0)
		assert(Fraction(1.0F, 2.0F).compareTo(Fraction.HALF) == 0)

		assert(Fraction(-1, -2).compareTo(Fraction.HALF) == 0)
		assert(Fraction(-1.0, -2.0).compareTo(Fraction.HALF) == 0)
		assert(Fraction(-1.0F, -2.0F).compareTo(Fraction.HALF) == 0)

		assert(Fraction(-1, 2).compareTo(Fraction.HALF) != 0)
		assert(Fraction(-1.0, 2.0).compareTo(Fraction.HALF) != 0)
		assert(Fraction(-1.0F, 2.0F).compareTo(Fraction.HALF) != 0)

		assert(Fraction(1, 2).compareTo(Fraction.ONE) != 0)
		assert(Fraction(1.0, 2.0).compareTo(Fraction.ONE) != 0)
		assert(Fraction(1.0F, 2.0F).compareTo(Fraction.ONE) != 0)

		assert(Fraction(2).compareTo(Fraction.TWO) == 0)
		assert(Fraction(2.0).compareTo(Fraction.TWO) == 0)
		assert(Fraction(2.0F).compareTo(Fraction.TWO) == 0)
		assert(Fraction(2, 1).compareTo(Fraction.TWO) == 0)
		assert(Fraction(2.0, 1.0).compareTo(Fraction.TWO) == 0)
		assert(Fraction(2.0F, 1.0F).compareTo(Fraction.TWO) == 0)

		assert(Fraction(4, 3) > Fraction(5, 4))
		assert(Fraction(4, 4) < Fraction(5, 4))

		assert(Fraction(-15, 2) < Fraction(2, 4))
		assert(Fraction(-1, 2) < Fraction(2, 4))
		assert(Fraction(-15, -2) > Fraction(2, 4))
		assert(Fraction(-15, -2) > Fraction(-2, -4))
		assert(Fraction(-15, 2) < Fraction(-2, -4))
	}

	@Test
	@DisplayName("Fraction math")
	fun math() {
		assert((Fraction(1) / Fraction(1)) == Fraction(1))
		assert((Fraction(2) / Fraction(1)) == Fraction(2, 1))
		assert((Fraction(2) / Fraction(2)) == Fraction(1))

		assert((Fraction(4, 3) / Fraction(5, 4)) == (Fraction(4, 3) * Fraction(4, 5)))
		assert((Fraction(4, 3) + Fraction(5, 4)) == Fraction(31, 12))
	}

	@Test
	@DisplayName("Fraction serialization test")
	fun serialization() {
		var value = Fraction(1)
		var serialized = value.serializeNBT()

		assert(value.compareTo(Fraction.deserializeNBT(serialized)) == 0)

		value = Fraction(4, 2)
		serialized = value.serializeNBT()

		assert(value.compareTo(Fraction.deserializeNBT(serialized)) == 0)

		value = Fraction(-4, 2)
		serialized = value.serializeNBT()

		assert(value.compareTo(Fraction.deserializeNBT(serialized)) == 0)

		value = Fraction(-4, -18)
		serialized = value.serializeNBT()

		assert(value.compareTo(Fraction.deserializeNBT(serialized)) == 0)

		value = Fraction("3407203485237459085739045724837543569234750927348902374590872345", "-57777772398450982374590230984532984")
		serialized = value.serializeNBT()

		assert(value.compareTo(Fraction.deserializeNBT(serialized)) == 0)
	}

	@Test
	@DisplayName("Fraction inaccurate representation")
	fun inaccurate() {
		var value = 1.1
		var frac = Fraction(value)

		assert(frac.toDouble() == value) { "$value != $frac as (${frac.toDouble()})" }

		value = 1.45
		frac = Fraction(value)

		assert(frac.toDouble() == value) { "$value != $frac as (${frac.toDouble()})" }

		value = -1.0
		frac = Fraction(value)

		assert(frac.toDouble() == value) { "$value != $frac as (${frac.toDouble()})" }

		value = -254.66
		frac = Fraction(value)

		assert(frac.toDouble() == value) { "$value != $frac as (${frac.toDouble()})" }

		value = 94.949494
		frac = Fraction(value)

		assert(frac.toDouble() == value) { "$value != $frac as (${frac.toDouble()})" }

		value = 1.0 / 3
		frac = Fraction(value)

		assert(frac.toDouble() == value) { "$value != $frac as (${frac.toDouble()})" }
	}
}
